# -*- coding: utf-8 -*-

"""The Tk graphical representation of an edge in the graph, i.e. an
arrow connecting nodes.

The information is stored in the graph object as attributes of the
edge so that the graphical representation can be restored as needed.
"""

import logging
import math
import seamm
import pprint  # noqa: F401
from tkinter import font
import tkinter as tk
import weakref

logger = logging.getLogger(__name__)


class TkEdge(seamm.Edge):
    str_to_object = weakref.WeakValueDictionary()

    def __init__(
        self,
        graph,
        node1,
        node2,
        edge_type='execution',
        edge_subtype='next',
        canvas=None,
        anchor1='s',
        anchor2='n',
        coords=None,
        **kwargs
    ):
        """Initialize the edge, ensuring that it is
        in the graph.

        Keyword arguments:
        """
        self._data = []
        logger.debug('Creating TkEdge {}'.format(self))
        logger.debug('\tnode1 = {}'.format(node1))
        logger.debug('\tnode2 = {}'.format(node2))

        # Initialize the parent class
        super().__init__(
            graph, node1, node2, edge_type, edge_subtype, **kwargs
        )

        self._data['canvas'] = canvas
        self.anchor1 = anchor1
        self.anchor2 = anchor2
        if coords is None:
            x0, y0 = self.node1.anchor_point(self.anchor1)
            x1, y1 = self.node2.anchor_point(self.anchor2)
            self.coords = [x0, y0, x1, y1]
        else:
            self.coords = coords

        # Remember the object so can get from tags on the canvas
        TkEdge.str_to_object[str(id(self))] = self

        # Arrange that the graphics are deleted when we are
        self._finalizer = weakref.finalize(
            self, self.canvas.delete, self.tag()
        )
        self._finalizer.atexit = False

    @property
    def canvas(self):
        return self._data['canvas']

    @property
    def anchor1(self):
        return self._data['anchor1']

    @anchor1.setter
    def anchor1(self, value):
        self._data['anchor1'] = value

    @property
    def anchor2(self):
        return self._data['anchor2']

    @anchor2.setter
    def anchor2(self, value):
        self._data['anchor2'] = value

    @property
    def coords(self):
        return self._data['coords']

    @coords.setter
    def coords(self, value):
        self._data['coords'] = value

    @property
    def has_label(self):
        return 'label_id' in self._data

    @property
    def label_id(self):
        return self._data['label_id']

    @property
    def label_bg_id(self):
        return self._data['label_bg_id']

    def tag(self):
        """Return a string tag for self"""
        return 'edge=' + str(id(self))

    def draw(self):
        """Draw the arrow for this edge"""
        self.move()

    def move(self):
        """Redraw the arrow when the nodes have moved"""

        x0, y0 = self.node1.anchor_point(self.anchor1)
        x1, y1 = self.node2.anchor_point(self.anchor2)
        self.coords[0] = x0
        self.coords[1] = y0
        self.coords[-2] = x1
        self.coords[-1] = y1

        # the arrow
        self.canvas.delete(self.tag() + '&& type=arrow')
        arrow_id = self.canvas.create_line(
            self.coords, arrow=tk.LAST, tags=[self.tag(), 'type=arrow']
        )
        self._data['arrow_id'] = arrow_id

        # and the label
        if self.edge_subtype != "next":
            self.canvas.delete(self.tag() + '&& type=label')
            text = self.canvas.create_text(
                self.label_position(x0, y0, x1, y1),
                text=self.edge_subtype,
                font=font.Font(family='Helvetica', size=8),
                tags=[self.tag(), 'type=label']
            )
            self._data['label_id'] = text
            self.canvas.delete(self.tag() + '&& type=label_bg')
            bg = self.canvas.create_rectangle(
                self.canvas.bbox(text),
                outline="white",
                fill="white",
                tags=[self.tag(), 'type=label_bg']
            )
            self._data['label_bg_id'] = bg
            self.canvas.tag_lower(bg, text)

    def label_position(self, x0, y0, x1, y1, offset=15):
        """Work out the position for the label on an edge"""
        dx = x1 - x0
        dy = y1 - y0
        length = math.sqrt(dx * dx + dy * dy)
        if length < 2 * offset:
            offset = int(length / 2)
        xy = [
            x0 if dx == 0.0 else x0 + dx / length * offset,
            y0 if dy == 0.0 else y0 + dy / length * offset
        ]
        return xy

    def undraw(self):
        """Remove any graphics"""
        self.canvas.delete(self.tag())
        if 'arrow_id' in self._data:
            del self._data['arrow_id']
        if 'label_id' in self._data:
            del self._data['label_id']
        if 'label_bg_id' in self._data:
            del self._data['label_bg_id']



# -*- coding: utf-8 -*-

"""The start node in a flowchart"""

import logging
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __  # noqa: F401

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('start')


class StartNode(seamm.Node):

    def __init__(self, flowchart=None):
        '''Initialize a specialized start node, which is the
        anchor for the graph.

        Keyword arguments:
        '''
        logger.debug('Constructing start node {}'.format(self))
        super().__init__(flowchart=flowchart, title='Start')

        self._uuid = 1

    def set_uuid(self):
        pass

    def describe(self, indent='', json_dict=None):
        """Write out information about what this node will do
        If json_dict is passed in, add information to that dictionary
        so that it can be written out by the controller as appropriate.
        """

        next_node = super().describe(indent, json_dict)

        return next_node

    def run(self):
        """'Run' the start node, i.e. do nothing but print
        """

        next_node = super().run(printer)
        return next_node

    def setup_printing(self, aprinter):
        """Establish the handlers for printing as controlled by
        options. The start step never writes to disk, so don't
        create that handler.
        """

        # First remove an existing handlers
        self.close_printing(printer)

        # A handler for stdout
        console_handler = logging.StreamHandler()
        console_handler.setLevel(printing.JOB)
        console_handler.setFormatter(self.formatter)
        printer.addHandler(console_handler)



# -*- coding: utf-8 -*-

"""The ExecWorkflow object does what it name implies: it executes, or
runs, a given flowchart.

It provides the environment for running the computational tasks
locally or remotely, using what is commonly called workflow management
system (WMS).  The WMS concept, as used here, means tools that run
given tasks without knowing anything about chemistry. The chemistry
specialization is contained in the Flowchart and the nodes that it
contains."""

import logging
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __  # noqa: F401

logger = logging.getLogger(__name__)
job = printing.getPrinter()


class ExecFlowchart(object):

    def __init__(self, flowchart=None):
        """Execute a flowchart, providing support for the actual
        execution of codes """

        self.flowchart = flowchart

    def run(self, root=None):
        if not self.flowchart:
            raise RuntimeError('There is no flowchart to run!')

        logger.debug('Creating global variables space')
        seamm.flowchart_variables = seamm.Variables()

        self.flowchart.root_directory = root

        # Correctly number the nodes
        self.flowchart.set_ids()

        # Write out an initial summary of the flowchart before doing anything
        # Reset the visited flag for traversal
        self.flowchart.reset_visited()

        # Get the start node
        next_node = self.flowchart.get_node('1')

        # describe ourselves
        job.job(
            (
                '\nDescription of the flowchart'
                '\n----------------------------'
            )
        )

        while next_node:
            next_node = next_node.describe()
        job.job('')

        # And actually run it!
        job.job(('Running the flowchart\n' '---------------------'))

        next_node = self.flowchart.get_node('1')
        while next_node:
            next_node = next_node.run()



# -*- coding: utf-8 -*-

import abc
import copy
import logging
import seamm
import pprint  # noqa: F401
import tkinter as tk
"""A graphical node using Tk on a canvas"""

logger = logging.getLogger(__name__)


class TkNode(abc.ABC):
    """The abstract base class for all Tk-based nodes"""

    anchor_points = {
        's': (+0.00, +0.50),
        'sse': (+0.25, +0.50),
        'se': (+0.50, +0.50),
        'ese': (+0.50, +0.25),
        'e': (+0.50, +0.00),
        'ene': (+0.50, -0.25),
        'ne': (+0.50, -0.50),
        'nne': (+0.25, -0.50),
        'n': (+0.00, -0.50),
        'nnw': (-0.25, -0.50),
        'nw': (-0.50, -0.50),
        'wnw': (-0.50, -0.25),
        'w': (-0.50, +0.00),
        'wsw': (-0.50, +0.25),
        'sw': (-0.50, +0.50),
        'ssw': (-0.25, +0.50)
    }

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=None,
        h=None
    ):
        """Initialize a node

        Keyword arguments:
        """
        self.tk_flowchart = tk_flowchart
        self.node = node
        self.toplevel = None
        self.canvas = canvas

        if self.node is not None:
            if self.node.x is None:
                self.node.x = x
            if self.node.y is None:
                self.node.y = y
            if self.node.w is None:
                self.node.w = w
            if self.node.h is None:
                self.node.h = h

        self.node_type = 'simple'

        self._border = None
        self.title_label = None
        self._selected = False
        self.popup_menu = None
        self._tmp = None
        self.dialog = None
        self.previous_grab = None

        # Widget information
        self._widget = {}
        self.tk_var = {}

    def __hash__(self):
        """Make iterable!"""
        return self.node.uuid

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets!"""
        return self._widget[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._widget[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self._widget:
            self._widget[key].destroy()
        del self._widget[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._widget)

    def __len__(self):
        """The len() command"""
        return len(self._widget)

    def __contains__(self, item):
        """Return a boolean indicating if a widget exists."""
        return item in self._widget

    @property
    def uuid(self):
        """The uuid of the node"""
        return self.node.uuid

    @property
    def title(self):
        """The title to display"""
        return self.node.title

    @title.setter
    def title(self, value):
        self.node.title = value
        if self.title_label is not None:
            self.canvas.itemconfigure(self.title_label, text=value)

    @property
    def tag(self):
        """The string representation of the uuid of the node"""
        return self.node.tag

    @property
    def flowchart(self):
        """The flowchart object"""
        return self.node.flowchart

    @flowchart.setter
    def flowchart(self, value):
        """The flowchart object"""
        self.node.flowchart = value

    @property
    def x(self):
        """The x-position of the center of the graphical node"""
        return self.node.x

    @x.setter
    def x(self, value):
        self.node.x = value

    @property
    def y(self):
        """The y-position of the center of the graphical node"""
        return self.node.y

    @y.setter
    def y(self, value):
        self.node.y = value

    @property
    def w(self):
        """The width of the graphical node"""
        return self.node.w

    @w.setter
    def w(self, value):
        self.node.w = value

    @property
    def h(self):
        """The height of the graphical node"""
        return self.node.h

    @h.setter
    def h(self, value):
        self.node.h = value

    def set_uuid(self):
        self.node.set_uuid()

    def connections(self):
        """Return a list of all the incoming and outgoing edges
        for this node, giving the anchor points and other node
        """

        return self.tk_flowchart.edges(self)

    @property
    def selected(self):
        """Whether I am selected or not"""
        return self._selected

    @selected.setter
    def selected(self, value):
        self._selected = value
        if value:
            self.canvas.itemconfigure(self.border, outline='red')
        else:
            self.canvas.itemconfigure(self.border, outline='black')

    @property
    def canvas(self):
        """The canvas for drawing the node"""
        return self._canvas

    @canvas.setter
    def canvas(self, value):
        if value:
            self.toplevel = value.winfo_toplevel()
        self._canvas = value

    @property
    def border(self):
        """The border of the picture in the flowchart"""
        return self._border

    @border.setter
    def border(self, value):
        self._border = value

    def draw(self):
        """Draw the node on the given canvas, making it visible"""

        # Remove any graphics items
        self.undraw()

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y - self.h / 2
        y1 = y0 + self.h
        self._border = self.canvas.create_rectangle(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, 'type=outline'],
            fill='white',
        )

        # the label in the middle
        self.title_label = self.canvas.create_text(
            self.x, self.y, text=self.title, tags=[self.tag, 'type=title']
        )

        for direction, edge in self.connections():
            edge.move()

    def undraw(self):
        """Remove all of our visual components
        """

        self.canvas.delete(self.tag)

    def move(self, deltax, deltay):
        if self._tmp is None:
            self._tmp = self.connections()

        self.x += deltax
        self.y += deltay

        self.canvas.move(self.tag, deltax, deltay)

        for connection in self._tmp:
            direction, edge = connection
            edge.move()

    def end_move(self, deltax, deltay):
        self.move(deltax, deltay)
        self._x0 = None
        self._y0 = None
        self._tmp = None

    def right_click(self, event):
        """Do whatever needs to be done for a right-click on this
        item in the flowchart.

        Subclasses should override this as appropriate! The menu
        created in this base method is accessible in subclasses
        which should make it relatively easy to override.
        """

        if self.popup_menu is not None:
            self.popup_menu.destroy()

        self.popup_menu = tk.Menu(self.canvas, tearoff=0)
        self.popup_menu.add_command(
            label="Delete",
            command=lambda: self.tk_flowchart.remove_node(self)
        )

        if type(self) is seamm.tk_node.TkNode:
            self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def double_click(self, event):
        """Do whatever needs to be done for a double-click on this
        item in the flowchart.

        Subclasses should override this as appropriate!
        """

        self.edit()

    def activate(self):
        """Add active handles at the anchor points and change the
        cursor
        """

        self.canvas.delete(self.tag + ' && type=anchor')
        for pt, x, y in self.anchor_point("all"):
            x0 = x - 2
            y0 = y - 2
            x1 = x + 2
            y1 = y + 2
            self.canvas.create_oval(
                x0,
                y0,
                x1,
                y1,
                fill='red',
                outline='red',
                tags=[self.tag, 'type=anchor', 'anchor=' + pt]
            )

    def deactivate(self):
        """Remove the decorations indicate the anchor points
        """

        self.canvas.delete(self.tag + ' && type=anchor')
        self.canvas.delete(self.tag + ' && type=active_anchor')

    def anchor_point(self, anchor="all"):
        """Where the anchor points are located. If "all" is given
        a dictionary of all points is returned"""

        if anchor == "all":
            result = []
            for pt in type(self).anchor_points:
                a, b = type(self).anchor_points[pt]
                result.append(
                    (pt, int(self.x + a * self.w), int(self.y + b * self.h))
                )
            return result

        if anchor in type(self).anchor_points:
            a, b = type(self).anchor_points[anchor]
            return (int(self.x + a * self.w), int(self.y + b * self.h))

        raise NotImplementedError(
            "anchor position '{}' not implemented".format(anchor)
        )

    def check_anchor_points(self, x, y, halo):
        """If the position x, y is within halo or one of the anchor points
        activate the point and return the name of the anchor point
        """

        points = []
        for direction, edge in self.connections():
            if direction == 'out':
                points.append(edge.anchor1)
            else:
                points.append(edge.anchor2)

        for point, x0, y0 in self.anchor_point():
            if x >= x0 - halo and x <= x0 + halo and \
               y >= y0 - halo and y <= y0 + halo:
                if point in points:
                    return None
                else:
                    return point
        return None

    def is_inside(self, x, y, halo=0):
        """Return a boolean indicating whether the point x, y is inside
        this node, using halo as a size around the point
        """
        if x < self.x - self.w / 2 - halo:
            return False
        if x > self.x + self.w / 2 + halo:
            return False

        if y < self.y - self.h / 2 - halo:
            return False
        if y > self.y + self.h / 2 + halo:
            return False

        return True

    def activate_anchor_point(self, point, halo):
        """Put a marker on the anchor point to indicate it is
        active
        """

        x, y = self.anchor_point(point)
        self.canvas.create_oval(
            x - halo,
            y - halo,
            x + halo,
            y + halo,
            fill='red',
            outline='red',
            tags=[self.tag, 'type=active_anchor', 'anchor=' + point]
        )

    def remove_edge(self, edge):
        """Remove a given edge, or all edges if 'all' is given
        """

        if isinstance(edge, str) and edge == 'all':
            for direction, obj in self.connections():
                self.remove_edge(obj)
        else:
            self.tk_flowchart.graph.remove_edge(
                edge.node1, edge.node2, edge.edge_type, edge.edge_subtype
            )

    def edit(self):
        """Do-nothing base class method"""
        pass

    def to_dict(self):
        """Serialize to a dict"""
        data = {
            'x': self._x,
            'y': self._y,
            'w': self._w,
            'h': self._h,
        }

        return data

    def update_flowchart(self, tk_flowchart=None, flowchart=None):
        """Update the nongraphical flowchart. Only used in nodes that contain
        flowcharts"""
        if tk_flowchart is None or flowchart is None:
            return

        # Make sure there is nothing in the flowchart
        flowchart.clear(all=True)

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in tk_flowchart:
            translate[node] = flowchart.add_node(copy.copy(node.node))
            node.update_flowchart()

        # And the edges
        for edge in tk_flowchart.edges():
            attr = {}
            for key in edge:
                if key not in (
                    'node1', 'node2', 'edge_type', 'edge_subtype', 'canvas'
                ):
                    attr[key] = edge[key]
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            flowchart.add_edge(
                node1, node2, edge.edge_type, edge.edge_subtype, **attr
            )

    def from_flowchart(self, tk_flowchart=None, flowchart=None):
        """Recreate the graphics from the non-graphical flowchart.
        Only used in nodes that contain flowchart"""

        if tk_flowchart is None or flowchart is None:
            return

        tk_flowchart.clear()

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in flowchart:
            extension = node.extension
            if extension is None:
                # Start node
                translate[node] = tk_flowchart.get_node('1')
            else:
                new_node = copy.copy(node)
                logger.debug('creating {} node'.format(extension))
                plugin = tk_flowchart.plugin_manager.get(extension)
                logger.debug('  plugin object: {}'.format(plugin))
                tk_node = plugin.create_tk_node(
                    tk_flowchart=tk_flowchart,
                    canvas=tk_flowchart.canvas,
                    node=new_node
                )
                translate[node] = tk_node
                tk_node.from_flowchart()
                tk_flowchart.graph.add_node(tk_node)
                tk_node.draw()

        # And the edges
        for edge in flowchart.edges():
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            attr = {}
            for key in edge:
                if key not in ('node1', 'node2'):
                    attr[key] = edge[key]
            tk_flowchart.add_edge(node1, node2, **attr)

    def default_edge_subtype(self):
        """Return the default subtype of the edge. Usually this is ''
        but for nodes with two or more edges leaving them, such as a loop, this
        method will return an appropriate default for the current edge. For
        example, by default the first edge emanating from a loop-node is the
        'loop' edge; the second, the 'exit' edge.

        A return value of 'too many' indicates that the node exceeds the number
        of allowed exit edges.
        """

        # how many outgoing edges are there?
        n_edges = len(self.tk_flowchart.edges(self, direction='out'))

        logger.debug('node.default_edge_label, n_edges = {}'.format(n_edges))

        if n_edges == 0:
            return "next"
        else:
            return "too many"

    def next_anchor(self):
        """Return where the next node should be positioned. The default is
        <gap> below the 's' anchor point.
        """

        return 's'



# -*- coding: utf-8 -*-

import collections.abc
import logging
import seamm
import pprint
"""A simple graph structure for holding the flowchart. This handles a
directed graph -- all edges have a direction implied -- with zero or
more edges from or to each node.
"""

logger = logging.getLogger(__name__)


class Graph(object):
    """A datastructure for holding a directed graph with multiple (parallel) edges.
    """

    def __init__(self):
        """Create the graph object"""

        self._node = {}
        self._edge = {}

    def __iter__(self):
        return self._node.values().__iter__()

    def __contains__(self, node):
        return node.__hash__() in self._node

    def add_node(self, node):
        if node in self:
            raise RuntimeError('node is already in the graph')
        self._node[node.__hash__()] = node
        return node

    def remove_node(self, node):
        if node not in self:
            raise RuntimeError('node is not in the graph')
        del self._node[node.__hash__()]

    def clear(self):
        self._node = {}
        self._edge = {}

    def add_edge(
        self,
        u,
        v,
        edge_type=None,
        edge_subtype=None,
        edge_class=None,
        **kwargs
    ):

        if u not in self:
            self.add_node(u)

        if v not in self:
            self.add_node(v)

        key = (u.__hash__(), v.__hash__(), edge_type, edge_subtype)
        if edge_class is None:
            self._edge[key] = seamm.Edge(
                self,
                u,
                v,
                edge_type=edge_type,
                edge_subtype=edge_subtype,
                **kwargs
            )
        else:
            self._edge[key] = edge_class(
                self,
                u,
                v,
                edge_type=edge_type,
                edge_subtype=edge_subtype,
                **kwargs
            )
        return self._edge[key]

    def remove_edge(self, u, v, edge_type=None, edge_subtype=None):
        key = (u.__hash__(), v.__hash__(), edge_type, edge_subtype)
        if key not in self._edge:
            raise RuntimeError('edge does not exist!')
        del self._edge[key]

    def edges(self, node=None, direction='both'):
        result = []
        if node is None:
            return self._edge.values()
        else:
            h = node.__hash__()
            if direction == 'both':
                for key in self._edge:
                    h1, h2, edge_type, edge_subtype = key
                    if h1 == h:
                        result.append(('out', self._edge[key]))
                    if h2 == h:
                        result.append(('in', self._edge[key]))
            elif direction == 'out':
                for key in self._edge:
                    h1, h2, edge_type, edge_subtype = key
                    if h1 == h:
                        result.append(self._edge[key])
            elif direction == 'in':
                for key in self._edge:
                    h1, h2, edge_type, edge_subtype = key
                    if h2 == h:
                        result.append(self._edge[key])
            else:
                return RuntimeError(
                    "Don't recognize direction '{}'!".format(direction)
                )

        return result

    def has_edge(self, u, v, edge_type=None, edge_subtype=None):
        key = (u.__hash__(), v.__hash__(), edge_type, edge_subtype)
        return key in self._edge


class Edge(collections.abc.MutableMapping):

    def __init__(
        self,
        graph,
        node1,
        node2,
        edge_type='execution',
        edge_subtype='next',
        **kwargs
    ):
        self.graph = graph
        self._data = dict(**kwargs)
        self._data['node1'] = node1
        self._data['node2'] = node2
        self._data['edge_type'] = edge_type
        self._data['edge_subtype'] = edge_subtype

    def __getitem__(self, key):
        """Allow [] access to the dictionary!"""
        return self._data[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._data[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        del self._data[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._data)

    def __len__(self):
        """The len() command"""
        return len(self._data)

    def __repr__(self):
        """The string representation of this object"""
        return repr(self._data)

    def __str__(self):
        """The pretty string representation of this object"""
        return pprint.pformat(self._data)

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        if item in self._data:
            return True
        return False

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self._data == other._data

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self._data.copy()

    @property
    def node1(self):
        return self._data['node1']

    @property
    def node2(self):
        return self._data['node2']

    @property
    def edge_type(self):
        return self._data['edge_type']

    @property
    def edge_subtype(self):
        return self._data['edge_subtype']


if __name__ == "__main__":

    class Node(object):

        def __init__(self, **kwargs):
            self.data = dict(**kwargs)

        def __str__(self):
            return pprint.pformat(self.data)

    graph = Graph()

    start = Node(title='start')
    node1 = Node(title='node1')
    node2 = Node(title='node2')

    edge1 = graph.add_edge(start, node1)
    print('edge1 = {}'.format(edge1))
    edge2 = graph.add_edge(node1, node2)
    print('edge2 = {}'.format(edge2))

    print('nodes:')
    for node in graph:
        print(' node = {}'.format(node))
        print()

    print('edges:')
    for edge in graph.edges():
        print(edge)
        print()



# -*- coding: utf-8 -*-

"""A node to join the flow in a flowchart"""

import logging
import seamm

logger = logging.getLogger(__name__)


class TkJoin(seamm.TkNode):
    """The Tk-based graphical representation of a joining node"""

    anchor_points = {
        'n': (0, -0.5),
        's': (0, 0.5),
        'e': (0.5, 0.0),
        'w': (-0.5, 0.0),
    }

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=120,
        y=20,
        w=10,
        h=10
    ):
        '''Initialize a node

        Keyword arguments:
        '''
        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h
        )

    def draw(self):
        """Draw the node on the given canvas, making it visible"""
        # Remove any graphics items
        self.undraw()

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y - self.h / 2
        y1 = y0 + self.h
        self._border = self.canvas.create_oval(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, 'type=outline'],
            fill='black',
        )

        for direction, edge in self.connections():
            edge.move()



from datetime import datetime
import argparse
import json
import locale
import logging
import seamm
import seamm_util  # MUST come after seamm
import seamm_util.printing as printing
import os
import os.path
import shutil
import sys

logger = logging.getLogger(__name__)
variables = seamm.Variables()


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def run():
    """The standalone flowchart app
    """

    parser = argparse.ArgumentParser(description='Execute a MolSSI flowchart')

    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose_count",
        action="count",
        default=0,
        help="increases log verbosity for each occurence."
    )
    parser.add_argument(
        "--directory",
        dest="directory",
        default=None,
        action="store",
        help="Directory to write output and other files."
    )
    parser.add_argument("--force", dest="force", action='store_true')
    parser.add_argument(
        "--output",
        choices=['files', 'stdout', 'both'],
        default='files',
        help='whether to put the output in files, direct to stdout, or both'
    )
    parser.add_argument("filename", help='the filename of the flowchart')

    args = parser.parse_args()

    # Set up logging level to WARNING by default, going more verbose
    # for each new -v, to INFO and then DEBUG and finally ALL with 3 -v's

    numeric_level = max(3 - args.verbose_count, 0) * 10
    logging.basicConfig(level=numeric_level)

    # Create the working directory where files, output, etc. go

    if args.directory is None:
        wdir = os.path.join(
            os.getcwd(),
            datetime.now().isoformat(sep='_', timespec='seconds')
        )
    else:
        wdir = args.directory

    logging.info("The working directory is '{}'".format(wdir))

    if os.path.exists(wdir):
        if args.force:
            shutil.rmtree(wdir)
        else:
            msg = "Directory '{}' exists, use --force to overwrite"\
                  .format(wdir)

            logging.critical(msg)
            sys.exit(msg)

    os.makedirs(wdir, exist_ok=False)

    # Set up the root printer, and add handlers to print to the file
    # 'job.out' in the working directory and to stdout, as requested
    # in the options. Since all printers are children of the root
    # printer, all output at the right levels will flow here

    printer = printing.getPrinter()

    # Set up our formatter
    formatter = logging.Formatter(fmt='{message:s}', style='{')

    # A handler for stdout
    console_handler = logging.StreamHandler()
    console_handler.setLevel(printing.JOB)
    console_handler.setFormatter(formatter)
    printer.addHandler(console_handler)

    # A handler for the file
    file_handler = logging.FileHandler(os.path.join(wdir, 'job.out'))
    file_handler.setLevel(printing.JOB)
    file_handler.setFormatter(formatter)
    printer.addHandler(file_handler)

    # And ... finally ... run!
    printer.job("Running in directory '{}'".format(wdir))

    # copy the flowchart to the root directory for later reference
    shutil.copy2(args.filename, os.path.join(wdir, 'flowchart.flow'))

    flowchart = seamm.Flowchart(directory=wdir, output=args.output)
    flowchart.read(args.filename)

    exec = seamm.ExecFlowchart(flowchart)
    with cd(wdir):
        exec.run(root=wdir)


def open_flowchart(name):
    with open(name, 'r') as fd:
        line = fd.readline(256)
        # There may be exec magic as first line
        if line[0:2] == '#!':
            line = fd.readline(256)
        if line[0:7] != '!MolSSI':
            raise RuntimeError('File is not a MolSSI file! -- ' + line)
        tmp = line.split()
        if len(tmp) < 3:
            raise RuntimeError('File is not a proper MolSSI file! -- ' + line)
        if tmp[1] != 'flowchart':
            raise RuntimeError('File is not a flowchart! -- ' + line)
        flowchart_version = tmp[2]
        logger.info(
            'Reading flowchart version {} from file {}'.format(
                flowchart_version, name
            )
        )

        data = json.load(fd, cls=seamm_util.JSONDecoder)

    if data['class'] != 'Flowchart':
        raise RuntimeError(
            'File {} does not contain a flowchart!'.format(name)
        )
        return

    # Restore the flowchart
    flowchart = seamm.Flowchart()
    flowchart.from_dict(data)

    return flowchart


if __name__ == "__main__":
    locale.setlocale(locale.LC_ALL, '')
    run()



# -*- coding: utf-8 -*-

import json
from datetime import datetime
import logging
import seamm
import seamm_util  # MUST come after seamm
import os
import os.path
import pprint  # noqa: F401
import stat
"""A flowchart, which is a set of nodes. There must be a single
'start' node, with other nodes connected via their ports to describe
the flowchart. There may be isolated nodes or groups of connected nodes;
however, the flow starts at the 'start' node and follows the connections,
so isolated nodes and fragments will not be executed."""

logger = logging.getLogger(__name__)


class Flowchart(object):
    """The class variable 'graphics' gives
    the default graphics to use for display, if needed. It defaults to
    'Tk' for the tkinter GUI.
    """

    graphics = 'Tk'

    def __init__(
        self,
        parent=None,
        data=None,
        namespace='org.molssi.seamm',
        name=None,
        directory=None,
        output='files'
    ):
        '''Initialize the flowchart

        Keyword arguments:
        '''

        self.graph = seamm.Graph()

        self.name = name
        self.parent = parent

        self.output = output  # Where to print output, files, stdout, both

        # Setup the plugin handling
        self.plugin_manager = seamm.PluginManager(namespace)

        # and make sure that the start node exists
        self.add_node(seamm.StartNode(flowchart=self))

        # And the root directory
        self.root_directory = directory

    def __iter__(self):
        return self.graph.__iter__()

    @property
    def root_directory(self):
        """The root directory for files, etc for this flowchart"""
        if self._root_directory is None:
            self._root_directory = os.path.join(
                os.getcwd(),
                datetime.now().isoformat(sep='_', timespec='seconds')
            )
        return self._root_directory

    @root_directory.setter
    def root_directory(self, value):
        self._root_directory = value

    @property
    def output(self):
        """Where to print output:
            files:  to files in subdirectories
            stdout: to standard output (for intereactive use)
            both:   to both files and standard output
        """
        return self._output

    @output.setter
    def output(self, value):
        if value in ('files', 'stdout', 'both'):
            self._output = value
        else:
            raise RuntimeError(
                "flowchart.output must be one of 'files', 'stdout', or 'both'"
                ", not '{}'".format(value)
            )

    # -------------------------------------------------------------------------
    # Node creation and deletion
    # -------------------------------------------------------------------------

    def create_node(self, extension_name):
        """Create a new node given the extension name"""
        plugin = self.plugin_manager.get(extension_name)
        node = plugin.create_node(flowchart=self, extension=extension_name)
        node.parent = self.parent
        return node

    def add_node(self, n, **attr):
        """Add a single node n, ensuring that it knows the flowchart"""
        n.flowchart = self
        return self.graph.add_node(n, **attr)

    def remove_node(self, node):
        """Delete a node from the flowchart, and from the graphics if
        necessary
        """

        # Remove the drawing of the node
        node.undraw()

        # and any edges, including graphics if appropriate
        node.remove_edge('all')

        # and the node
        self.graph.remove_node(node)

    # -------------------------------------------------------------------------
    # Finding nodes
    # -------------------------------------------------------------------------

    def tag_exists(self, tag):
        """Check if the node with a given tag exists. A tag is a string like
        'node=<uuid>', where <uuid> is the unique integer id for the node.
        """

        for node in self:
            if node.tag == tag:
                return True
        return False

    def get_node(self, uuid):
        """Return the node with a given uuid"""
        if isinstance(uuid, int):
            uuid = str(uuid)
        for node in self:
            if str(node.uuid) == uuid:
                return node
        return None

    def last_node(self, node='1'):
        """Find the last node walking down the main execution path
        from the given node, which defaults to the start node"""

        if isinstance(node, str):
            node = self.get_node(node)

        for edge in self.graph.edges(node, direction='out'):
            if edge.edge_type == "execution":
                return self.last_node(edge.node2)

        return node

    def clear(self, all=False):
        """Override the underlying clear() to ensure that the start node is present
        """
        self.graph.clear()

        # and make sure that the start node exists
        if not all:
            start_node = seamm.StartNode(flowchart=self)
            self.add_node(start_node)

    def list_nodes(self):
        """List the nodes, for debugging"""
        result = []
        for node in self:
            result.append(node.__class__.__name__ + " {}".format(node))
        return result

    # -------------------------------------------------------------------------
    # Taversal
    # -------------------------------------------------------------------------

    def reset_visited(self):
        """Reset the 'visited' flag, which is used to detect
        loops during traversals
        """

        for tmp in self:
            tmp.visited = False

    def set_ids(self, node_id=()):
        """Sequentially number all nodes, and subnodes"""
        logger.debug('Setting ids')

        # Clear all ids
        for node in self:
            node.reset_id()

        # Reset the visited flag to check for loops
        self.reset_visited()

        # Get the start node
        next_node = self.get_node('1')

        # And traverse the nodes.
        n = 0
        while next_node:
            next_node = next_node.set_id((*node_id, str(n)))
            n += 1
        logger.debug('Finished setting ids')

    # -------------------------------------------------------------------------
    # Strings, reading and writing
    # -------------------------------------------------------------------------

    def to_json(self):
        """Ufff. Turn ourselves into JSON"""
        return json.dumps(self.to_dict())

    def to_dict(self):
        """Serialize the graph and everything it contains in a dict"""

        data = {
            'item': 'object',
            'module': self.__module__,
            'class': self.__class__.__name__,
            'extension': None
        }

        nodes = data['nodes'] = []
        for node in self:
            nodes.append(node.to_dict())

        edges = data['edges'] = []
        for edge in self.graph.edges():
            attr = {}
            for key in edge:
                if key not in ('node1', 'node2', 'edge_type', 'edge_subtype'):
                    attr[key] = edge[key]
            edges.append(
                {
                    'item': 'edge',
                    'node1': edge.node1.uuid,
                    'node2': edge.node2.uuid,
                    'edge_type': edge.edge_type,
                    'edge_subtype': edge.edge_subtype,
                    'attributes': attr
                }
            )

        return data

    def from_dict(self, data):
        """recreate the flowchart from the dict. Inverse of to_dict."""
        if 'class' not in data:
            raise RuntimeError('There is no class information in the data!')
        if data['class'] != self.__class__.__name__:
            raise RuntimeError(
                'The dictionary does not contain a flowchart!'
                ' It contains a {} class'.format(data['class'])
            )

        self.clear()

        # Recreate the nodes
        for node in data['nodes']:
            if node['class'] == 'StartNode':
                new_node = self.get_node('1')
                new_node.from_dict(node)
                continue

            logger.debug('creating {} node'.format(node['extension']))
            plugin = self.plugin_manager.get(node['extension'])
            logger.debug('  plugin object: {}'.format(plugin))

            # Recreate the node
            new_node = plugin.create_node(
                flowchart=self, extension=node['extension']
            )
            new_node.parent = self.parent

            # set uuid to correct value
            new_node._uuid = node['attributes']['_uuid']

            # and add to the flowchart
            self.add_node(new_node)

            new_node.from_dict(node)

            logger.debug(
                "adding nodes: nodes:\n\t" + "\n\t".join(self.list_nodes())
            )

        # and the edges connecting them
        for edge in data['edges']:
            node1 = self.get_node(edge['node1'])
            node2 = self.get_node(edge['node2'])
            self.add_edge(
                node1,
                node2,
                edge_type=edge['edge_type'],
                edge_subtype=edge['edge_subtype'],
                **edge['attributes']
            )

            logger.debug(
                "Adding edges, nodes:\n\t" + "\n\t".join(self.list_nodes())
            )

    def write(self, filename):
        """Write the serialized form to disk"""
        with open(filename, 'w') as fd:
            fd.write('#!/usr/bin/env run_flowchart\n')
            fd.write('!MolSSI flowchart 1.0\n')
            json.dump(self.to_dict(), fd, indent=4, cls=seamm_util.JSONEncoder)
            logger.info('Wrote json to {}'.format(filename))

        permissions = stat.S_IMODE(os.lstat(filename).st_mode)
        os.chmod(filename, permissions | stat.S_IXUSR | stat.S_IXGRP)

    def read(self, filename):
        """Recreate the flowchart from the serialized form on disk"""
        with open(filename, 'r') as fd:
            line = fd.readline(256)
            # There may be exec magic as first line
            if line[0:2] == '#!':
                line = fd.readline(256)
            if line[0:7] != '!MolSSI':
                raise RuntimeError('File is not a MolSSI file! -- ' + line)
            tmp = line.split()
            if len(tmp) < 3:
                raise RuntimeError(
                    'File is not a proper MolSSI file! -- ' + line
                )
            if tmp[1] != 'flowchart':
                raise RuntimeError('File is not a flowchart! -- ' + line)
            flowchart_version = tmp[2]
            logger.info(
                'Reading flowchart version {} from file {}'.format(
                    flowchart_version, filename
                )
            )

            data = json.load(fd, cls=seamm_util.JSONDecoder)

        if data['class'] != 'Flowchart':
            raise RuntimeError(filename + ' does not contain a flowchart')

        self.from_dict(data)

    # -------------------------------------------------------------------------
    # Edges between nodes
    # -------------------------------------------------------------------------

    def edges(self, node=None, direction='both'):
        return self.graph.edges(node, direction)

    def add_edge(self, u, v, edge_type=None, edge_subtype='next', **attr):
        return self.graph.add_edge(u, v, edge_type, edge_subtype, **attr)

    def print_edges(self, event=None):
        '''Print all the edges. Useful for debugging!
        '''

        print('All edges in flowchart')
        for edge in self.edges():
            # print('   {}'.format(edge))
            print(
                '   {} {} {} {} {} {}'.format(
                    edge.node1.tag, edge.anchor1, edge.node2.tag, edge.anchor2,
                    edge.edge_type, edge.edge_subtype
                )
            )

    # -------------------------------------------------------------------------
    # Printing
    # -------------------------------------------------------------------------



# -*- coding: utf-8 -*-

import collections.abc
import logging
import seamm
import pprint
"""A dictionary-like object for holding variables accessible to the
executing flowchart.
"""

logger = logging.getLogger(__name__)
# Module level variable to be used in other modules
flowchart_variables = None


class Variables(collections.abc.MutableMapping):

    def __init__(self, **kwargs):
        self._data = dict(**kwargs)

    def __getitem__(self, key):
        """Allow [] access to the dictionary!"""
        return self._data[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._data[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        del self._data[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._data)

    def __len__(self):
        """The len() command"""
        return len(self._data)

    def __repr__(self):
        """The string representation of this object"""
        return repr(self._data)

    def __str__(self):
        """The pretty string representation of this object"""
        return pprint.pformat(self._data)

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        if item in self._data:
            return True
        return False

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self._data == other._data

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self._data.copy()

    def value(self, string):
        """Return the value of the variable or expression if it is an
        expression, i.e. starts with a $ and optionally has braces around the
        variable name.

        If it is not a variable, return the original string unchanged
        """

        if isinstance(string, str) and string[0] == '$':
            expression = self.filter_expression(string)

            result = eval(expression, seamm.flowchart_variables._data)
            return result
        else:
            return string

    def set_variable(self, variable, value):
        """Set the value of the variable. The variable may be a simple string
        or start with a $ and optionally have braces around it, i.e.

            <name>
            $<name>
        or
            ${<name>}
        """

        name = self.variable(variable)
        self._data[name] = value

    def get_variable(self, variable):
        """Get the value of the variable. The variable may be a simple string
        or start with a $ and optionally have braces around it, i.e.

            <name>
            $<name>
        or
            ${<name>}
        """

        name = self.variable(variable)
        if name not in self._data:
            raise RuntimeError(
                "Workspace variable '{}' does not exist.".format(name)
            )
        return self._data[name]

    def exists(self, variable):
        """Return whether a variable exists. The variable may be specified
        as a simple string or start with a $ and optionally have braces
        around it, i.e.

            <name>
            $<name>
        or
            ${<name>}
        """

        return self.variable(variable) in self._data

    def delete(self, variable):
        """Return whether a variable exists. The variable may be specified
        as a simple string or start with a $ and optionally have braces
        around it, i.e.

            <name>
            $<name>
        or
            ${<name>}
        """

        name = self.variable(variable)
        if name in self._data:
            del self._data[name]

    def variable(self, string):
        """Return the name of a variable. The variable may be specified
        as a simple string or start with a $ and optionally have braces
        around it, i.e.

            <string>
            $<string>
        or
            ${<string>}
        """

        if string[0] == '$':
            if string[1] == '{':
                if string[-1] != '}':
                    raise RuntimeError(
                        "'" + string + "'is not a valid string name"
                    )
                else:
                    return string[2:-1]
            else:
                return string[1:]
        else:
            return string

    def filter_expression(self, string):
        """A variable or expression coming from the GUI uses '$'
        to indicate a variable, optionally bracketing the variable
        with braces, i.e. ${name}.

        This method filters out the variable markers, respecting
        quoted string, returning a string that can be eval'ed in
        the Python interpreter.
        """

        result = ''
        state = ''
        for char in string:
            if state == 'in single quotes':
                result += char
                if char == "'":
                    state = ''
            elif state == 'in double quotes':
                result += char
                if char == '"':
                    state = ''
            elif state == 'protected':
                result += char
                state = ''
            elif state == 'in variable name':
                if char == '}':
                    state = ''
                elif char == '{':
                    pass
                else:
                    result += char
            else:
                if char == '$':
                    state = 'in variable name'
                elif char == "'":
                    result += char
                    state = 'in single quotes'
                elif char == '"':
                    result += char
                    state = 'in double quotes'
                elif char == '\\':
                    state = 'protected'
                else:
                    result += char

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pillow',
    'Pmw',
    'argparse',
    'bitmath',
    'datetime',
    'seamm_util',
    'numpy',
    'pandas',
    'pint',
    'pprint',
    'pyuca',
    'stevedore',
]

setup_requirements = [
    'pytest-runner',
    # TODO(paulsaxe): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    'pytest',
    # TODO: put package test requirements here
]

setup(
    name='seamm',
    version='0.1.0',
    description="Simulation Environment for Atomistic and Molecular Modeling ",
    long_description=readme + '\n\n' + history,
    author="Paul Saxe",
    author_email='psaxe@molssi.org',
    url='https://github.com/molssi-seamm/seamm',
    packages=find_packages(include=['seamm']),
    entry_points={
        'console_scripts': [
            'seamm=seamm.__main__:flowchart',
            'run_flowchart=seamm.run_flowchart:run'
        ],
        'org.molssi.seamm': [
            'Split = seamm:SplitStep',
        ],
        'org.molssi.seamm.tk': [
            'Split = seamm:SplitStep',
        ],
        'org.molssi.seamm': [
            'Join = seamm:JoinStep',
        ],
        'org.molssi.seamm.tk': [
            'Join = seamm:JoinStep',
        ],
    },
    include_package_data=True,
    install_requires=requirements,
    license="GNU Lesser General Public License v3",
    zip_safe=False,
    keywords='seamm',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



import argparse
import seamm
import locale
import logging
import os
import platform
import Pmw
import subprocess
import sys
import tkinter as tk

logger = logging.getLogger(__name__)
dbg_level = 30


def raise_app(root: tk):
    root.attributes("-topmost", True)
    if platform.system() == 'Darwin':
        tmpl = (
            'tell application "System Events" to set frontmost '
            'of every process whose unix id is {} to true'
        )
        script = tmpl.format(os.getpid())
        subprocess.check_call(['/usr/bin/osascript', '-e', script])
    root.after(100, lambda: root.attributes("-topmost", False))


def flowchart():
    """The standalone flowchart app
    """
    global app_name
    app_name = 'MolSSI Framework'
    global dbg_level

    parser = argparse.ArgumentParser(description='MolSSI Framework')
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose_count",
        action="count",
        default=0,
        help="increases log verbosity for each occurence."
    )
    parser.add_argument(
        "flowcharts",
        nargs='*',
        default=[],
        help='flowcharts to open initially'
    )

    args = parser.parse_args()

    # Sets log level to WARN going more verbose for each new -v.
    dbg_level = max(3 - args.verbose_count, 0) * 10
    logging.basicConfig(level=dbg_level)

    ##################################################
    # Initialize Tk
    ##################################################
    root = tk.Tk()
    Pmw.initialise(root)

    ##############################################################
    # Create the various objects that we need: the model, the view
    # and the data
    ##############################################################

    framework = seamm.Flowchart()
    tk_framework = seamm.TkFlowchart(master=root, flowchart=framework)
    # The data is implicitly initialized to none...

    ##################################################
    # Initialize the rest of the GUI, such as menus
    ##################################################
    logger.debug('Initializing the rest of the GUI')

    root.title(app_name)

    # The menus
    menu = tk.Menu(root)

    # Set the about and preferences menu items on Mac
    if sys.platform.startswith('darwin'):
        app_menu = tk.Menu(menu, name='apple')
        menu.add_cascade(menu=app_menu)

        app_menu.add_command(
            label='About ' + app_name, command=tk_framework.about
        )
        app_menu.add_separator()
        root.createcommand(
            'tk::mac::ShowPreferences', tk_framework.preferences
        )
        root.createcommand('tk::mac::OpenDocument', tk_framework.open_file)
        CmdKey = 'Command-'
    else:
        CmdKey = 'Control-'

    root.config(menu=menu)
    filemenu = tk.Menu(menu)
    menu.add_cascade(label="File", menu=filemenu)
    filemenu.add_command(
        label="New", command=tk_framework.new_file, accelerator=CmdKey + 'N'
    )
    filemenu.add_command(
        label="Save...", command=tk_framework.save, accelerator=CmdKey + 'S'
    )
    filemenu.add_command(label="Save as...", command=tk_framework.save_file)
    filemenu.add_command(
        label="Open...",
        command=tk_framework.open_file,
        accelerator=CmdKey + 'O'
    )
    filemenu.add_separator()
    filemenu.add_command(
        label="Run", command=tk_framework.run, accelerator=CmdKey + 'R'
    )

    # Control debugging info
    filemenu.add_separator()
    debug_menu = tk.Menu(menu)
    filemenu.add_cascade(label="Debug", menu=debug_menu)
    debug_menu.add_radiobutton(
        label='normal',
        value=30,
        variable=dbg_level,
        command=lambda arg0=30: handle_dbg_level(arg0)
    )
    debug_menu.add_radiobutton(
        label='info',
        value=20,
        variable=dbg_level,
        command=lambda arg0=20: handle_dbg_level(arg0)
    )
    debug_menu.add_radiobutton(
        label='debug',
        value=10,
        variable=dbg_level,
        command=lambda arg0=10: handle_dbg_level(arg0)
    )

    # Exiting
    filemenu.add_separator()
    filemenu.add_command(label="Exit", command=root.quit)

    # Edit menu
    editmenu = tk.Menu(menu)
    menu.add_cascade(label="Edit", menu=editmenu)
    filemenu.add_command(
        label="Clean layout",
        command=tk_framework.clean_layout,
        accelerator=CmdKey + 'C'
    )

    # Help menu
    helpmenu = tk.Menu(menu)
    menu.add_cascade(label="Help", menu=helpmenu)
    root.createcommand('tk::mac::ShowHelp', tk_framework.help)
    root.bind_all('<' + CmdKey + 'N>', tk_framework.new_file)
    root.bind_all('<' + CmdKey + 'n>', tk_framework.new_file)
    root.bind_all('<' + CmdKey + 'O>', tk_framework.open_file)
    root.bind_all('<' + CmdKey + 'o>', tk_framework.open_file)
    root.bind_all('<' + CmdKey + 'R>', tk_framework.run)
    root.bind_all('<' + CmdKey + 'r>', tk_framework.run)
    root.bind_all('<' + CmdKey + 'S>', tk_framework.save)
    root.bind_all('<' + CmdKey + 's>', tk_framework.save)
    root.bind_all('<' + CmdKey + 'C>', tk_framework.clean_layout)
    root.bind_all('<' + CmdKey + 'c>', tk_framework.clean_layout)

    root.bind_all('<' + CmdKey + '`>', tk_framework.print_edges)

    # Work out and set the window size to nicely fit the screen
    sw = root.winfo_screenwidth()
    sh = root.winfo_screenheight()
    w = int(0.9 * sw)
    h = int(0.8 * sh)
    x = int(0.1 * sw / 2)
    y = int(0.2 * sh / 2)

    root.geometry('{}x{}+{}+{}'.format(w, h, x, y))

    logger.debug('Finished initializing the rest of the GUI, drawing window')

    # Draw the flowchart
    tk_framework.draw()

    logger.debug('Framework has been drawn. Now raise it to the top')

    # bring it to the top of all windows
    root.lift()
    raise_app(root)

    # Check to see if the command line has flowcharts to open
    if len(args.flowcharts) > 0:
        logger.debug('open the following flowcharts:')
        if len(args.flowcharts) > 1:
            raise RuntimeError('Currently handle only one flowchart at a time')
        for filename in args.flowcharts:
            tk_framework.open(filename)

    logger.debug('and now enter the event loop')

    # enter the event loop
    root.mainloop()


def handle_dbg_level(level):
    global dbg_level

    dbg_level = level
    logging.getLogger().setLevel(dbg_level)


if __name__ == "__main__":
    locale.setlocale(locale.LC_ALL, '')

    flowchart()



# -*- coding: utf-8 -*-

"""A node to split the flow into parallel segements in a flowchart"""

import seamm
import logging

logger = logging.getLogger(__name__)


class Split(seamm.Node):

    def __init__(self, flowchart=None, extension=None):
        '''Initialize a node for splitting the flow apart

        Keyword arguments:
        '''
        logger.debug('Constructing split node {}'.format(self))
        super().__init__(flowchart=flowchart, title='Split')



# -*- coding: utf-8 -*-

"""A node to join together parallel flows in a flowchart"""

import seamm
import logging

logger = logging.getLogger(__name__)


class Join(seamm.Node):

    def __init__(self, flowchart=None, extension='Join'):
        '''Initialize a node for joining the flow together again

        Keyword arguments:
        '''
        logger.debug('Constructing join node {}'.format(self))
        super().__init__(
            flowchart=flowchart, title='Join', extension=extension
        )



# -*- coding: utf-8 -*-

"""Helper class needed for the stevedore integration. Needs to provide
a description() method that returns a dict containing a description of
this node, and a factory() method for creating the graphical and non-graphical
nodes."""

import seamm


class SplitStep(object):
    my_description = {
        'description': 'An interface for a node to split the control flow',
        'group': 'Control',
        'name': 'Split'
    }

    def __init__(self, flowchart=None, gui=None):
        """Initialize this helper class, which is used by
        the application via stevedore to get information about
        and create node objects for the flowchart
        """
        pass

    def description(self):
        """Return a description of what this extension does
        """
        return SplitStep.my_description

    def create_node(self, flowchart=None, **kwargs):
        """Return the new node object"""
        return seamm.Split(flowchart=flowchart, **kwargs)

    def create_tk_node(self, canvas=None, **kwargs):
        """Return the graphical Tk node object"""
        return seamm.TkSplit(canvas=canvas, **kwargs)


class JoinStep(object):
    my_description = {
        'description': 'An interface for a node to join the control flow',
        'group': 'Control',
        'name': 'Join'
    }

    def __init__(self, flowchart=None, gui=None):
        """Initialize this helper class, which is used by
        the application via stevedore to get information about
        and create node objects for the flowchart
        """
        pass

    def description(self):
        """Return a description of what this extension does
        """
        return JoinStep.my_description

    def create_node(self, flowchart=None, **kwargs):
        """Return the new node object"""
        return seamm.Join(flowchart=flowchart, **kwargs)

    def create_tk_node(self, canvas=None, **kwargs):
        """Return the graphical Tk node object"""
        return seamm.TkJoin(canvas=canvas, **kwargs)



# -*- coding: utf-8 -*-

"""Top-level package for SEAMM."""

import textwrap

__author__ = """Paul Saxe"""
__email__ = 'psaxe@molssi.org'
__version__ = '0.1.0'

# Text handling
from textwrap import dedent  # noqa: F401

# Bring up the classes so that they appear to be directly in
# the seamm package.

from seamm.parameters import Parameter  # noqa: F401
from seamm.parameters import Parameters  # noqa: F401
from seamm.variables import Variables  # noqa: F401
from seamm.variables import flowchart_variables  # noqa: F401
from seamm.plugin_manager import PluginManager  # noqa: F401
from seamm.flowchart import Flowchart  # noqa: F401
from seamm.tk_flowchart import TkFlowchart  # noqa: F401
from seamm.graph import Graph  # noqa: F401
from seamm.graph import Edge  # noqa: F401
from seamm.exec_flowchart import ExecFlowchart  # noqa: F401
from seamm.node import Node  # noqa: F401
from seamm.start_node import StartNode  # noqa: F401
from seamm.tk_edge import TkEdge  # noqa: F401
from seamm.tk_node import TkNode  # noqa: F401
from seamm.tk_start_node import TkStartNode  # noqa: F401
from seamm.exec_local import ExecLocal  # noqa: F401
from seamm.split_node import Split  # noqa: F401
from seamm.tk_split_node import TkSplit  # noqa: F401
from seamm.builtins import SplitStep  # noqa: F401
from seamm.join_node import Join  # noqa: F401
from seamm.tk_join_node import TkJoin  # noqa: F401
from seamm.builtins import JoinStep  # noqa: F401

wrap_text = textwrap.TextWrapper(width=120)
wrap_stdout = textwrap.TextWrapper(width=120)
